@@ -11,58 +11,42 @@
     Implements support for high-level dataset access.
 """
 
-from __future__ import absolute_import
-
+try:
+    from functools import cached_property
+except ImportError:
+    from cached_property import cached_property
 import posixpath as pp
 import sys
 
 from threading import local
 
-import six
-from six.moves import xrange    # pylint: disable=redefined-builtin
-
 import numpy
 
-from .. import h5, h5s, h5t, h5r, h5d, h5p, h5fd
-from .base import HLObject, phil, with_phil, Empty, is_empty_dataspace
+from .. import h5, h5s, h5t, h5r, h5d, h5p, h5fd, h5ds, _selector
+from .base import HLObject, phil, with_phil, Empty, find_item_type
 from . import filters
 from . import selections as sel
 from . import selections2 as sel2
 from .datatype import Datatype
+from .compat import filename_decode
+from .vds import VDSmap, vds_support
 
 _LEGACY_GZIP_COMPRESSION_VALS = frozenset(range(10))
 MPI = h5.get_config().mpi
 
 
-def readtime_dtype(basetype, names):
-    """ Make a NumPy dtype appropriate for reading """
-
-    if len(names) == 0:  # Not compound, or we want all fields
-        return basetype
-
-    if basetype.names is None:  # Names provided, but not compound
-        raise ValueError("Field names only allowed for compound types")
-
-    for name in names:  # Check all names are legal
-        if not name in basetype.names:
-            raise ValueError("Field %s does not appear in this type." % name)
-
-    return numpy.dtype([(name, basetype.fields[name][0]) for name in names])
-
-
-def make_new_dset(parent, shape=None, dtype=None, data=None,
-                 chunks=None, compression=None, shuffle=None,
-                    fletcher32=None, maxshape=None, compression_opts=None,
-                  fillvalue=None, scaleoffset=None, track_times=None):
-    """ Return a new low-level dataset identifier
-
-    Only creates anonymous datasets.
-    """
+def make_new_dset(parent, shape=None, dtype=None, data=None, name=None,
+                  chunks=None, compression=None, shuffle=None,
+                  fletcher32=None, maxshape=None, compression_opts=None,
+                  fillvalue=None, scaleoffset=None, track_times=None,
+                  external=None, track_order=None, dcpl=None,
+                  allow_unknown_filter=False):
+    """ Return a new low-level dataset identifier """
 
     # Convert data to a C-contiguous ndarray
     if data is not None and not isinstance(data, Empty):
         from . import base
-        data = numpy.asarray(data, order="C", dtype=base.guess_dtype(data))
+        data = base.array_for_new_object(data, specified_dtype=dtype)
 
     # Validate shape
     if shape is None:
@@ -72,14 +56,19 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
             data = Empty(dtype)
         shape = data.shape
     else:
-        shape = tuple(shape)
-        if data is not None and (numpy.product(shape) != numpy.product(data.shape)):
+        shape = (shape,) if isinstance(shape, int) else tuple(shape)
+        if data is not None and (numpy.product(shape, dtype=numpy.ulonglong) != numpy.product(data.shape, dtype=numpy.ulonglong)):
             raise ValueError("Shape tuple is incompatible with data")
 
+    if isinstance(maxshape, int):
+        maxshape = (maxshape,)
     tmp_shape = maxshape if maxshape is not None else shape
+
     # Validate chunk shape
+    if isinstance(chunks, int) and not isinstance(chunks, bool):
+        chunks = (chunks,)
     if isinstance(chunks, tuple) and any(
-        chunk > dim for dim, chunk in zip(tmp_shape,chunks) if dim is not None
+        chunk > dim for dim, chunk in zip(tmp_shape, chunks) if dim is not None
     ):
         errmsg = "Chunk shape must not be greater than data shape in any dimension. "\
                  "{} is not compatible with {}".format(chunks, shape)
@@ -100,7 +89,7 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
         tid = h5t.py_create(dtype, logical=1)
 
     # Legacy
-    if any((compression, shuffle, fletcher32, maxshape,scaleoffset)) and chunks is False:
+    if any((compression, shuffle, fletcher32, maxshape, scaleoffset)) and chunks is False:
         raise ValueError("Chunked format required for given storage options")
 
     # Legacy
@@ -115,9 +104,10 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
             raise TypeError("Conflict in compression options")
         compression_opts = compression
         compression = 'gzip'
-
-    dcpl = filters.generate_dcpl(shape, dtype, chunks, compression, compression_opts,
-                  shuffle, fletcher32, maxshape, scaleoffset)
+    dcpl = filters.fill_dcpl(
+        dcpl or h5p.create(h5p.DATASET_CREATE), shape, dtype,
+        chunks, compression, compression_opts, shuffle, fletcher32,
+        maxshape, scaleoffset, external, allow_unknown_filter)
 
     if fillvalue is not None:
         fillvalue = numpy.array(fillvalue)
@@ -127,6 +117,13 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
         dcpl.set_obj_track_times(track_times)
     elif track_times is not None:
         raise TypeError("track_times must be either True or False")
+    if track_order is True:
+        dcpl.set_attr_creation_order(
+            h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
+    elif track_order is False:
+        dcpl.set_attr_creation_order(0)
+    elif track_order is not None:
+        raise TypeError("track_order must be either True or False")
 
     if maxshape is not None:
         maxshape = tuple(m if m is not None else h5s.UNLIMITED for m in maxshape)
@@ -137,7 +134,7 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
         sid = h5s.create_simple(shape, maxshape)
 
 
-    dset_id = h5d.create(parent.id, None, tid, sid, dcpl=dcpl)
+    dset_id = h5d.create(parent.id, name, tid, sid, dcpl=dcpl)
 
     if (data is not None) and (not isinstance(data, Empty)):
         dset_id.write(h5s.ALL, h5s.ALL, data)
@@ -145,25 +142,109 @@ def make_new_dset(parent, shape=None, dtype=None, data=None,
     return dset_id
 
 
-class AstypeContext(object):
+def make_new_virtual_dset(parent, shape, sources, dtype, name=None,
+                          maxshape=None, fillvalue=None):
+    """ Return a new low-level dataset identifier for a virtual dataset """
 
-    """
-        Context manager which allows changing the type read from a dataset.
-    """
+    # create the creation property list
+    dcpl = h5p.create(h5p.DATASET_CREATE)
+    if fillvalue is not None:
+        dcpl.set_fill_value(numpy.array([fillvalue]))
 
+    if maxshape is not None:
+        maxshape = tuple(m if m is not None else h5s.UNLIMITED for m in maxshape)
+
+    virt_dspace = h5s.create_simple(shape, maxshape)
+
+    for vspace, fpath, dset, src_dspace in sources:
+        dcpl.set_virtual(vspace, fpath, dset, src_dspace)
+
+    if isinstance(dtype, Datatype):
+        # Named types are used as-is
+        tid = dtype.id
+    else:
+        dtype = numpy.dtype(dtype)
+        tid = h5t.py_create(dtype, logical=1)
+
+    return h5d.create(parent.id, name=name, tid=tid, space=virt_dspace,
+                      dcpl=dcpl)
+
+
+class AstypeWrapper(object):
+    """Wrapper to convert data on reading from a dataset.
+    """
     def __init__(self, dset, dtype):
         self._dset = dset
         self._dtype = numpy.dtype(dtype)
 
+    def __getitem__(self, args):
+        return self._dset.__getitem__(args, new_dtype=self._dtype)
+
     def __enter__(self):
         # pylint: disable=protected-access
         self._dset._local.astype = self._dtype
+        return self
 
     def __exit__(self, *args):
         # pylint: disable=protected-access
         self._dset._local.astype = None
 
 
+class AsStrWrapper:
+    """Wrapper to decode strings on reading the dataset"""
+    def __init__(self, dset, encoding, errors='strict'):
+        self.dset = dset
+        if encoding is None:
+            encoding = h5t.check_string_dtype(dset.dtype).encoding
+        self.encoding = encoding
+        self.errors = errors
+
+    def __getitem__(self, args):
+        bytes_arr = self.dset[args]
+        # numpy.char.decode() seems like the obvious thing to use. But it only
+        # accepts numpy string arrays, not object arrays of bytes (which we
+        # return from HDF5 variable-length strings). And the numpy
+        # implementation is not faster than doing it with a loop; in fact, by
+        # not converting the result to a numpy unicode array, the
+        # naive way can be faster! (Comparing with numpy 1.18.4, June 2020)
+        if numpy.isscalar(bytes_arr):
+            return bytes_arr.decode(self.encoding, self.errors)
+
+        return numpy.array([
+            b.decode(self.encoding, self.errors) for b in bytes_arr.flat
+        ], dtype=object).reshape(bytes_arr.shape)
+
+
+class FieldsWrapper:
+    """Wrapper to extract named fields from a dataset with a struct dtype"""
+    extract_field = None
+
+    def __init__(self, dset, prior_dtype, names):
+        self.dset = dset
+        if isinstance(names, str):
+            self.extract_field = names
+            names = [names]
+        self.read_dtype = readtime_dtype(prior_dtype, names)
+
+    def __getitem__(self, args):
+        data = self.dset.__getitem__(args, new_dtype=self.read_dtype)
+        if self.extract_field is not None:
+            data = data[self.extract_field]
+        return data
+
+
+def readtime_dtype(basetype, names):
+    """Make a NumPy compound dtype with a subset of available fields"""
+    if basetype.names is None:  # Names provided, but not compound
+        raise ValueError("Field names only allowed for compound types")
+
+    for name in names:  # Check all names are legal
+        if not name in basetype.names:
+            raise ValueError("Field %s does not appear in this type." % name)
+
+    return numpy.dtype([(name, basetype.fields[name][0]) for name in names])
+
+
 if MPI:
     class CollectiveContext(object):
 
@@ -183,6 +264,79 @@ if MPI:
             self._dset._dxpl.set_dxpl_mpio(h5fd.MPIO_INDEPENDENT)
 
 
+class ChunkIterator(object):
+    """
+    Class to iterate through list of chunks of a given dataset
+    """
+    def __init__(self, dset, source_sel=None):
+        self._shape = dset.shape
+        rank = len(dset.shape)
+
+        if not dset.chunks:
+            # can only use with chunked datasets
+            raise TypeError("Chunked dataset required")
+
+        self._layout = dset.chunks
+        if source_sel is None:
+            # select over entire dataset
+            slices = []
+            for dim in range(rank):
+                slices.append(slice(0, self._shape[dim]))
+            self._sel = tuple(slices)
+        else:
+            if isinstance(source_sel, slice):
+                self._sel = (source_sel,)
+            else:
+                self._sel = source_sel
+        if len(self._sel) != rank:
+            raise ValueError("Invalid selection - selection region must have same rank as dataset")
+        self._chunk_index = []
+        for dim in range(rank):
+            s = self._sel[dim]
+            if s.start < 0 or s.stop > self._shape[dim] or s.stop <= s.start:
+                raise ValueError("Invalid selection - selection region must be within dataset space")
+            index = s.start // self._layout[dim]
+            self._chunk_index.append(index)
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        rank = len(self._shape)
+        slices = []
+        if rank == 0 or self._chunk_index[0] * self._layout[0] >= self._sel[0].stop:
+            # ran past the last chunk, end iteration
+            raise StopIteration()
+
+        for dim in range(rank):
+            s = self._sel[dim]
+            start = self._chunk_index[dim] * self._layout[dim]
+            stop = (self._chunk_index[dim] + 1) * self._layout[dim]
+            # adjust the start if this is an edge chunk
+            if start < s.start:
+                start = s.start
+            if stop > s.stop:
+                stop = s.stop  # trim to end of the selection
+            s = slice(start, stop, 1)
+            slices.append(s)
+
+        # bump up the last index and carry forward if we run outside the selection
+        dim = rank - 1
+        while dim >= 0:
+            s = self._sel[dim]
+            self._chunk_index[dim] += 1
+
+            chunk_end = self._chunk_index[dim] * self._layout[dim]
+            if chunk_end < s.stop:
+                # we still have room to extend along this dimensions
+                return tuple(slices)
+
+            if dim > 0:
+                # reset to the start and continue iterating with higher dimension
+                self._chunk_index[dim] = 0
+            dim -= 1
+        return tuple(slices)
+
 class Dataset(HLObject):
 
     """
@@ -190,13 +344,44 @@ class Dataset(HLObject):
     """
 
     def astype(self, dtype):
-        """ Get a context manager allowing you to perform reads to a
+        """ Get a wrapper allowing you to perform reads to a
         different destination type, e.g.:
 
-        >>> with dataset.astype('f8'):
-        ...     double_precision = dataset[0:100:2]
+        >>> double_precision = dataset.astype('f8')[0:100:2]
         """
-        return AstypeContext(self, dtype)
+        return AstypeWrapper(self, dtype)
+
+    def asstr(self, encoding=None, errors='strict'):
+        """Get a wrapper to read string data as Python strings:
+
+        >>> str_array = dataset.asstr()[:]
+
+        The parameters have the same meaning as in ``bytes.decode()``.
+        If ``encoding`` is unspecified, it will use the encoding in the HDF5
+        datatype (either ascii or utf-8).
+        """
+        string_info = h5t.check_string_dtype(self.dtype)
+        if string_info is None:
+            raise TypeError(
+                "dset.asstr() can only be used on datasets with "
+                "an HDF5 string datatype"
+            )
+        if encoding is None:
+            encoding = string_info.encoding
+        return AsStrWrapper(self, encoding, errors=errors)
+
+    def fields(self, names, *, _prior_dtype=None):
+        """Get a wrapper to read a subset of fields from a compound data type:
+
+        >>> 2d_coords = dataset.fields(['x', 'y'])[:]
+
+        If names is a string, a single field is extracted, and the resulting
+        arrays will have that dtype. Otherwise, it should be an iterable,
+        and the read data will have a compound dtype.
+        """
+        if _prior_dtype is None:
+            _prior_dtype = self.dtype
+        return FieldsWrapper(self, _prior_dtype, names)
 
     if MPI:
         @property
@@ -219,10 +404,20 @@ class Dataset(HLObject):
         return self.id.rank
 
     @property
-    @with_phil
     def shape(self):
         """Numpy-style shape tuple giving dataset dimensions"""
-        return self.id.shape
+        if 'shape' in self._cache_props:
+            return self._cache_props['shape']
+
+        with phil:
+            shape = self.id.shape
+
+        # If the file is read-only, cache the shape to speed-up future uses.
+        # This cache is invalidated by .refresh() when using SWMR.
+        if self._readonly:
+            self._cache_props['shape'] = shape
+        return shape
+
     @shape.setter
     @with_phil
     def shape(self, shape):
@@ -230,10 +425,57 @@ class Dataset(HLObject):
         self.resize(shape)
 
     @property
-    @with_phil
     def size(self):
         """Numpy-style attribute giving the total dataset size"""
-        return numpy.prod(self.shape, dtype=numpy.intp)
+        if 'size' in self._cache_props:
+            return self._cache_props['size']
+
+        if self._is_empty:
+            size = None
+        else:
+            size = numpy.prod(self.shape, dtype=numpy.intp)
+
+        # If the file is read-only, cache the size to speed-up future uses.
+        # This cache is invalidated by .refresh() when using SWMR.
+        if self._readonly:
+            self._cache_props['size'] = size
+        return size
+
+    @property
+    def nbytes(self):
+        """Numpy-style attribute giving the raw dataset size as the number of bytes"""
+        size = self.size
+        if size is None:  # if we are an empty 0-D array, then there are no bytes in the dataset
+            return 0
+        return self.dtype.itemsize * size
+
+    @property
+    def _selector(self):
+        """Internal object for optimised selection of data"""
+        if '_selector' in self._cache_props:
+            return self._cache_props['_selector']
+
+        slr = _selector.Selector(self.id.get_space())
+
+        # If the file is read-only, cache the reader to speed up future uses.
+        # This cache is invalidated by .refresh() when using SWMR.
+        if self._readonly:
+            self._cache_props['_selector'] = slr
+        return slr
+
+    @property
+    def _fast_reader(self):
+        """Internal object for optimised reading of data"""
+        if '_fast_reader' in self._cache_props:
+            return self._cache_props['_fast_reader']
+
+        rdr = _selector.Reader(self.id)
+
+        # If the file is read-only, cache the reader to speed up future uses.
+        # This cache is invalidated by .refresh() when using SWMR.
+        if self._readonly:
+            self._cache_props['_fast_reader'] = rdr
+        return rdr
 
     @property
     @with_phil
@@ -241,14 +483,6 @@ class Dataset(HLObject):
         """Numpy dtype representing the datatype"""
         return self.id.dtype
 
-    @property
-    @with_phil
-    def value(self):
-        """  Alias for dataset[()] """
-        DeprecationWarning("dataset.value has been deprecated. "
-            "Use dataset[()] instead.")
-        return self[()]
-
     @property
     @with_phil
     def chunks(self):
@@ -297,6 +531,21 @@ class Dataset(HLObject):
         except KeyError:
             return None
 
+    @property
+    @with_phil
+    def external(self):
+        """External file settings. Returns a list of tuples of
+        (name, offset, size) for each external file entry, or returns None
+        if no external files are used."""
+        count = self._dcpl.get_external_count()
+        if count<=0:
+            return None
+        ext_list = list()
+        for x in range(count):
+            (name, offset, size) = self._dcpl.get_external(x)
+            ext_list.append( (filename_decode(name), offset, size) )
+        return ext_list
+
     @property
     @with_phil
     def maxshape(self):
@@ -304,6 +553,9 @@ class Dataset(HLObject):
         None have no resize limit. """
         space = self.id.get_space()
         dims = space.get_simple_extent_dims(True)
+        if dims is None:
+            return None
+
         return tuple(x if x != h5s.UNLIMITED else None for x in dims)
 
     @property
@@ -314,17 +566,30 @@ class Dataset(HLObject):
         self._dcpl.get_fill_value(arr)
         return arr[0]
 
+    @cached_property
     @with_phil
-    def __init__(self, bind):
+    def _extent_type(self):
+        """Get extent type for this dataset - SIMPLE, SCALAR or NULL"""
+        return self.id.get_space().get_simple_extent_type()
+
+    @cached_property
+    def _is_empty(self):
+        """Check if extent type is empty"""
+        return self._extent_type == h5s.NULL
+
+    @with_phil
+    def __init__(self, bind, *, readonly=False):
         """ Create a new Dataset object by binding to a low-level DatasetID.
         """
         if not isinstance(bind, h5d.DatasetID):
             raise ValueError("%s is not a DatasetID" % bind)
-        HLObject.__init__(self, bind)
+        super(Dataset, self).__init__(bind)
 
         self._dcpl = self.id.get_create_plist()
         self._dxpl = h5p.create(h5p.DATASET_XFER)
         self._filters = filters.get_filters(self._dcpl)
+        self._readonly = readonly
+        self._cache_props = {}
         self._local = local()
         self._local.astype = None
 
@@ -392,12 +657,36 @@ class Dataset(HLObject):
         shape = self.shape
         if len(shape) == 0:
             raise TypeError("Can't iterate over a scalar dataset")
-        for i in xrange(shape[0]):
+        for i in range(shape[0]):
             yield self[i]
 
+    @with_phil
+    def iter_chunks(self, sel=None):
+        """ Return chunk iterator.  If set, the sel argument is a slice or
+        tuple of slices that defines the region to be used. If not set, the
+        entire dataspace will be used for the iterator.
+
+        For each chunk within the given region, the iterator yields a tuple of
+        slices that gives the intersection of the given chunk with the
+        selection area.
+
+        A TypeError will be raised if the dataset is not chunked.
+
+        A ValueError will be raised if the selection region is invalid.
+
+        """
+        return ChunkIterator(self, sel)
+
+    @cached_property
+    def _fast_read_ok(self):
+        """Is this dataset suitable for simple reading"""
+        return (
+            self._extent_type == h5s.SIMPLE
+            and isinstance(self.id.get_type(), (h5t.TypeIntegerID, h5t.TypeFloatID))
+        )
 
     @with_phil
-    def __getitem__(self, args):
+    def __getitem__(self, args, new_dtype=None):
         """ Read a slice from the HDF5 dataset.
 
         Takes slices and recarray-style field names (more than one is
@@ -409,24 +698,35 @@ class Dataset(HLObject):
         * Boolean "mask" array indexing
         """
         args = args if isinstance(args, tuple) else (args,)
-        if is_empty_dataspace(self.id):
-            if not (args == tuple() or args == (Ellipsis,)):
-                raise ValueError("Empty datasets cannot be sliced")
-            return Empty(self.dtype)
 
-        # Sort field indices from the rest of the args.
-        names = tuple(x for x in args if isinstance(x, six.string_types))
-        args = tuple(x for x in args if not isinstance(x, six.string_types))
-        if six.PY2:
-            names = tuple(x.encode('utf-8') if isinstance(x, six.text_type) else x for x in names)
+        if new_dtype is None:
+            new_dtype = getattr(self._local, 'astype', None)
 
-        new_dtype = getattr(self._local, 'astype', None)
-        if new_dtype is not None:
-            new_dtype = readtime_dtype(new_dtype, names)
-        else:
-            # This is necessary because in the case of array types, NumPy
-            # discards the array information at the top level.
-            new_dtype = readtime_dtype(self.id.dtype, names)
+        if self._fast_read_ok and (new_dtype is None):
+            try:
+                return self._fast_reader.read(args)
+            except TypeError:
+                pass  # Fall back to Python read pathway below
+
+        if self._is_empty:
+            # Check 'is Ellipsis' to avoid equality comparison with an array:
+            # array equality returns an array, not a boolean.
+            if args == () or (len(args) == 1 and args[0] is Ellipsis):
+                return Empty(self.dtype)
+            raise ValueError("Empty datasets cannot be sliced")
+
+        # Sort field names from the rest of the args.
+        names = tuple(x for x in args if isinstance(x, str))
+
+        if names:
+            # Read a subset of the fields in this structured dtype
+            if len(names) == 1:
+                names = names[0]  # Read with simpler dtype of this field
+            args = tuple(x for x in args if not isinstance(x, str))
+            return self.fields(names, _prior_dtype=new_dtype)[args]
+
+        if new_dtype is None:
+            new_dtype = self.dtype
         mtype = h5t.py_create(new_dtype)
 
         # === Special-case region references ====
@@ -440,10 +740,12 @@ class Dataset(HLObject):
             sid = h5r.get_region(args[0], self.id)
             mshape = sel.guess_shape(sid)
             if mshape is None:
-                return numpy.array((0,), dtype=new_dtype)
-            if numpy.product(mshape) == 0:
-                return numpy.array(mshape, dtype=new_dtype)
+                # 0D with no data (NULL or deselected SCALAR)
+                return Empty(new_dtype)
             out = numpy.empty(mshape, dtype=new_dtype)
+            if out.size == 0:
+                return out
+
             sid_out = h5s.create_simple(mshape)
             sid_out.select_all()
             self.id.read(sid_out, sid, out, mtype)
@@ -451,9 +753,10 @@ class Dataset(HLObject):
 
         # === Check for zero-sized datasets =====
 
-        if numpy.product(self.shape) == 0:
-            # These are the only access methods NumPy allows for such objects
-            if args == (Ellipsis,) or args == tuple():
+        if self.size == 0:
+            # Check 'is Ellipsis' to avoid equality comparison with an array:
+            # array equality returns an array, not a boolean.
+            if args == () or (len(args) == 1 and args[0] is Ellipsis):
                 return numpy.empty(self.shape, dtype=new_dtype)
 
         # === Scalar dataspaces =================
@@ -464,8 +767,6 @@ class Dataset(HLObject):
             arr = numpy.ndarray(selection.mshape, dtype=new_dtype)
             for mspace, fspace in selection:
                 self.id.read(mspace, fspace, arr, mtype)
-            if len(names) == 1:
-                arr = arr[names[0]]
             if selection.mshape is None:
                 return arr[()]
             return arr
@@ -473,38 +774,23 @@ class Dataset(HLObject):
         # === Everything else ===================
 
         # Perform the dataspace selection.
-        selection = sel.select(self.shape, args, dsid=self.id)
+        selection = sel.select(self.shape, args, dataset=self)
 
         if selection.nselect == 0:
-            return numpy.ndarray(selection.mshape, dtype=new_dtype)
+            return numpy.ndarray(selection.array_shape, dtype=new_dtype)
 
-        # Up-converting to (1,) so that numpy.ndarray correctly creates
-        # np.void rows in case of multi-field dtype. (issue 135)
-        single_element = selection.mshape == ()
-        mshape = (1,) if single_element else selection.mshape
-        arr = numpy.ndarray(mshape, new_dtype, order='C')
+        arr = numpy.ndarray(selection.array_shape, new_dtype, order='C')
 
-        # HDF5 has a bug where if the memory shape has a different rank
-        # than the dataset, the read is very slow
-        if len(mshape) < len(self.shape):
-            # pad with ones
-            mshape = (1,)*(len(self.shape)-len(mshape)) + mshape
-
-        # Perfom the actual read
-        mspace = h5s.create_simple(mshape)
+        # Perform the actual read
+        mspace = h5s.create_simple(selection.mshape)
         fspace = selection.id
         self.id.read(mspace, fspace, arr, mtype, dxpl=self._dxpl)
 
         # Patch up the output for NumPy
-        if len(names) == 1:
-            arr = arr[names[0]]     # Single-field recarray convention
         if arr.shape == ():
-            arr = numpy.asscalar(arr)
-        if single_element:
-            arr = arr[0]
+            return arr[()]   # 0 dim array -> numpy scalar
         return arr
 
-
     @with_phil
     def __setitem__(self, args, val):
         """ Write to the HDF5 dataset from a Numpy array.
@@ -516,15 +802,13 @@ class Dataset(HLObject):
         args = args if isinstance(args, tuple) else (args,)
 
         # Sort field indices from the slicing
-        names = tuple(x for x in args if isinstance(x, six.string_types))
-        args = tuple(x for x in args if not isinstance(x, six.string_types))
-        if six.PY2:
-            names = tuple(x.encode('utf-8') if isinstance(x, six.text_type) else x for x in names)
+        names = tuple(x for x in args if isinstance(x, str))
+        args = tuple(x for x in args if not isinstance(x, str))
 
         # Generally we try to avoid converting the arrays on the Python
         # side.  However, for compound literals this is unavoidable.
-        vlen = h5t.check_dtype(vlen=self.dtype)
-        if vlen is not None and vlen not in (bytes, six.text_type):
+        vlen = h5t.check_vlen_dtype(self.dtype)
+        if vlen is not None and vlen not in (bytes, str):
             try:
                 val = numpy.asarray(val, dtype=vlen)
             except ValueError:
@@ -537,7 +821,7 @@ class Dataset(HLObject):
                 if val.ndim > 1:
                     tmp = numpy.empty(shape=val.shape[:-1], dtype=object)
                     tmp.ravel()[:] = [i for i in val.reshape(
-                        (numpy.product(val.shape[:-1]), val.shape[-1]))]
+                        (numpy.product(val.shape[:-1], dtype=numpy.ulonglong), val.shape[-1]))]
                 else:
                     tmp = numpy.array([None], dtype=object)
                     tmp[0] = val
@@ -560,8 +844,25 @@ class Dataset(HLObject):
             if cast_compound:
                 val = val.view(numpy.dtype([(names[0], dtype)]))
                 val = val.reshape(val.shape[:len(val.shape) - len(dtype.shape)])
+        elif (self.dtype.kind == 'S'
+              and (h5t.check_string_dtype(self.dtype).encoding == 'utf-8')
+              and (find_item_type(val) is str)
+        ):
+            # Writing str objects to a fixed-length UTF-8 string dataset.
+            # Numpy's normal conversion only handles ASCII characters, but
+            # when the destination is UTF-8, we want to allow any unicode.
+            # This *doesn't* handle numpy fixed-length unicode data ('U' dtype),
+            # as HDF5 has no equivalent, and converting fixed length UTF-32
+            # to variable length UTF-8 would obscure what's going on.
+            str_array = numpy.asarray(val, order='C', dtype=object)
+            val = numpy.array([
+                s.encode('utf-8') for s in str_array.flat
+            ], dtype=self.dtype).reshape(str_array.shape)
         else:
-            val = numpy.asarray(val, order='C')
+            # If the input data is already an array, let HDF5 do the conversion.
+            # If it's a list or similar, don't make numpy guess a dtype for it.
+            dt = None if isinstance(val, numpy.ndarray) else self.dtype
+            val = numpy.asarray(val, order='C', dtype=dt)
 
         # Check for array dtype compatibility and convert
         if self.dtype.subdtype is not None:
@@ -606,28 +907,40 @@ class Dataset(HLObject):
             mtype = None
 
         # Perform the dataspace selection
-        selection = sel.select(self.shape, args, dsid=self.id)
+        selection = sel.select(self.shape, args, dataset=self)
 
         if selection.nselect == 0:
             return
 
         # Broadcast scalars if necessary.
-        if mshape == () and selection.mshape != ():
+        # In order to avoid slow broadcasting filling the destination by
+        # the scalar value, we create an intermediate array of the same
+        # size as the destination buffer provided that size is reasonable.
+        # We assume as reasonable a size smaller or equal as the used dataset
+        # chunk size if any.
+        # In case of dealing with a non-chunked destination dataset or with
+        # a selection whose size is larger than the dataset chunk size we fall
+        # back to using an intermediate array of size equal to the last dimension
+        # of the destination buffer.
+        # The reasoning behind is that it makes sense to assume the creator of
+        # the dataset used an appropriate chunk size according the available
+        # memory. In any case, if we cannot afford to create an intermediate
+        # array of the same size as the dataset chunk size, the user program has
+        # little hope to go much further. Solves h5py issue #1067
+        if mshape == () and selection.array_shape != ():
             if self.dtype.subdtype is not None:
                 raise TypeError("Scalar broadcasting is not supported for array dtypes")
-            val2 = numpy.empty(selection.mshape[-1], dtype=val.dtype)
+            if self.chunks and (numpy.prod(self.chunks, dtype=numpy.float) >=
+                                numpy.prod(selection.array_shape, dtype=numpy.float)):
+                val2 = numpy.empty(selection.array_shape, dtype=val.dtype)
+            else:
+                val2 = numpy.empty(selection.array_shape[-1], dtype=val.dtype)
             val2[...] = val
             val = val2
             mshape = val.shape
 
         # Perform the write, with broadcasting
-        # Be careful to pad memory shape with ones to avoid HDF5 chunking
-        # glitch, which kicks in for mismatched memory/file selections
-        if len(mshape) < len(self.shape):
-            mshape_pad = (1,)*(len(self.shape)-len(mshape)) + mshape
-        else:
-            mshape_pad = mshape
-        mspace = h5s.create_simple(mshape_pad, (h5s.UNLIMITED,)*len(mshape_pad))
+        mspace = h5s.create_simple(selection.expand_shape(mshape))
         for fspace in selection.broadcast(mshape):
             self.id.write(mspace, fspace, val, mtype, dxpl=self._dxpl)
 
@@ -640,18 +953,18 @@ class Dataset(HLObject):
         Broadcasting is supported for simple indexing.
         """
         with phil:
-            if is_empty_dataspace(self.id):
+            if self._is_empty:
                 raise TypeError("Empty datasets have no numpy representation")
             if source_sel is None:
                 source_sel = sel.SimpleSelection(self.shape)
             else:
-                source_sel = sel.select(self.shape, source_sel, self.id)  # for numpy.s_
+                source_sel = sel.select(self.shape, source_sel, self)  # for numpy.s_
             fspace = source_sel.id
 
             if dest_sel is None:
                 dest_sel = sel.SimpleSelection(dest.shape)
             else:
-                dest_sel = sel.select(dest.shape, dest_sel, self.id)
+                dest_sel = sel.select(dest.shape, dest_sel, self)
 
             for mspace in dest_sel.broadcast(source_sel.mshape):
                 self.id.read(mspace, fspace, dest, dxpl=self._dxpl)
@@ -665,18 +978,18 @@ class Dataset(HLObject):
         Broadcasting is supported for simple indexing.
         """
         with phil:
-            if is_empty_dataspace(self.id):
+            if self._is_empty:
                 raise TypeError("Empty datasets cannot be written to")
             if source_sel is None:
                 source_sel = sel.SimpleSelection(source.shape)
             else:
-                source_sel = sel.select(source.shape, source_sel, self.id)  # for numpy.s_
+                source_sel = sel.select(source.shape, source_sel, self)  # for numpy.s_
             mspace = source_sel.id
 
             if dest_sel is None:
                 dest_sel = sel.SimpleSelection(self.shape)
             else:
-                dest_sel = sel.select(self.shape, dest_sel, self.id)
+                dest_sel = sel.select(self.shape, dest_sel, self)
 
             for fspace in dest_sel.broadcast(source_sel.mshape):
                 self.id.write(mspace, fspace, source, dxpl=self._dxpl)
@@ -684,13 +997,13 @@ class Dataset(HLObject):
     @with_phil
     def __array__(self, dtype=None):
         """ Create a Numpy array containing the whole dataset.  DON'T THINK
-        THIS MEANS DATASETS ARE INTERCHANGABLE WITH ARRAYS.  For one thing,
-        you have to read the whole dataset everytime this method is called.
+        THIS MEANS DATASETS ARE INTERCHANGEABLE WITH ARRAYS.  For one thing,
+        you have to read the whole dataset every time this method is called.
         """
         arr = numpy.empty(self.shape, dtype=self.dtype if dtype is None else dtype)
 
         # Special case for (0,)*-shape datasets
-        if numpy.product(self.shape) == 0:
+        if numpy.product(self.shape, dtype=numpy.ulonglong) == 0:
             return arr
 
         self.read_direct(arr)
@@ -699,18 +1012,16 @@ class Dataset(HLObject):
     @with_phil
     def __repr__(self):
         if not self:
-            r = u'<Closed HDF5 dataset>'
+            r = '<Closed HDF5 dataset>'
         else:
             if self.name is None:
-                namestr = u'("anonymous")'
+                namestr = '("anonymous")'
             else:
                 name = pp.basename(pp.normpath(self.name))
-                namestr = u'"%s"' % (name if name != u'' else u'/')
-            r = u'<HDF5 dataset %s: shape %s, type "%s">' % (
+                namestr = '"%s"' % (name if name != '' else '/')
+            r = '<HDF5 dataset %s: shape %s, type "%s">' % (
                 namestr, self.shape, self.dtype.str
             )
-        if six.PY2:
-            return r.encode('utf8')
         return r
 
     if hasattr(h5d.DatasetID, "refresh"):
@@ -719,9 +1030,10 @@ class Dataset(HLObject):
             """ Refresh the dataset metadata by reloading from the file.
 
             This is part of the SWMR features and only exist when the HDF5
-            librarary version >=1.9.178
+            library version >=1.9.178
             """
             self._id.refresh()
+            self._cache_props.clear()
 
     if hasattr(h5d.DatasetID, "flush"):
         @with_phil
@@ -730,6 +1042,38 @@ class Dataset(HLObject):
             If the dataset is chunked, raw data chunks are written to the file.
 
             This is part of the SWMR features and only exist when the HDF5
-            librarary version >=1.9.178
+            library version >=1.9.178
             """
             self._id.flush()
+
+    if vds_support:
+        @property
+        @with_phil
+        def is_virtual(self):
+            """Check if this is a virtual dataset"""
+            return self._dcpl.get_layout() == h5d.VIRTUAL
+
+        @with_phil
+        def virtual_sources(self):
+            """Get a list of the data mappings for a virtual dataset"""
+            if not self.is_virtual:
+                raise RuntimeError("Not a virtual dataset")
+            dcpl = self._dcpl
+            return [
+                VDSmap(dcpl.get_virtual_vspace(j),
+                       dcpl.get_virtual_filename(j),
+                       dcpl.get_virtual_dsetname(j),
+                       dcpl.get_virtual_srcspace(j))
+                for j in range(dcpl.get_virtual_count())]
+
+    @with_phil
+    def make_scale(self, name=''):
+        """Make this dataset an HDF5 dimension scale.
+
+        You can then attach it to dimensions of other datasets like this::
+
+            other_ds.dims[0].attach_scale(ds)
+
+        You can optionally pass a name to associate with this scale.
+        """
+        h5ds.set_scale(self._id, self._e(name))
